package no.ntnu.idatt1002.demo.controller;

import java.io.IOException;

import java.util.Optional;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Node;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.ComboBox;
import javafx.scene.control.Dialog;
import javafx.scene.control.TableColumn;
import javafx.scene.control.TableView;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.scene.text.Text;
import javafx.stage.Modality;
import javafx.stage.Stage;
import no.ntnu.idatt1002.demo.data.Economics.Income;
import no.ntnu.idatt1002.demo.data.Economics.IncomeCategory;
import no.ntnu.idatt1002.demo.data.Economics.IncomeRegister;
import no.ntnu.idatt1002.demo.data.Economics.FileHandling;

/**
 * Controller for income scene in the application. This controller manages all actions that relates to the income tableview (add, edit and delete), the switching
 * of scenes from the income scene to another scene, and the saving of the table, whenever the user switches to another scene.
 *
 * @author Harry Linrui Xu
 * @since 24.3.2023
 */
public class IncomeController {

  @FXML
  private Button addBtn;
  @FXML
  private Button editBtn;

  @FXML
  private Button deleteBtn;

  @FXML
  private ComboBox<String> show;

  @FXML
  private Button expenseBtn;

  @FXML
  private Button overviewBtn;

  @FXML
  private Button budgetBtn;
  @FXML
  private Button returnBtn;

  @FXML
  private Text sum;
  @FXML
  private TableColumn<Income, Double> amountColumn;

  @FXML
  private TableColumn<Income, IncomeCategory> categoryColumn;

  @FXML
  private TableColumn<Income, String> dateColumn;

  @FXML
  private TableColumn<Income, String> descriptionColumn;

  @FXML
  private TableColumn<Income, Boolean> recurringColumn;

  @FXML
  private TableView<Income> incomeTableView;

  IncomeRegister incomeRegister;
  ObservableList<Income> income;

  ObservableList<String> filter;

  /**
   * Initializes the income register, the observable income list and the tableview, along values of the dropbox used for filtering the tableview.
   * The method is called each time the FXML of this scene is loaded.
   * @throws IOException If there occurs any exception when loading the budget register from a file.
   */
  @FXML
  public void initialize() throws IOException {
    //Initialize table columns
    dateColumn.setCellValueFactory(new PropertyValueFactory<Income, String>("date"));
    amountColumn.setCellValueFactory(new PropertyValueFactory<Income, Double>("amount"));
    categoryColumn.setCellValueFactory(new PropertyValueFactory<Income, IncomeCategory>("category"));
    descriptionColumn.setCellValueFactory(new PropertyValueFactory<Income, String>("description"));
    recurringColumn.setCellValueFactory(new PropertyValueFactory<Income, Boolean>("recurring"));

    //Initialize the filter box
    filter = FXCollections.observableArrayList("All", "Gift", "Salary", "Student loan", "Fixed income");
    show.setItems(filter);
    show.setValue("All");

    //Initialize registers and tableview
    incomeRegister = loadIncomeDataFromFile("Income");
    income = FXCollections.observableArrayList(incomeRegister.getItems());
    incomeTableView.setItems(income);

    //Initialize sum field under the tableview
    sum.setText(String.valueOf(incomeRegister.getTotalSum()));
    //if budget.register isEmpty -> disable expense
  }

  /**
   * Method for handling the adding of new entries in the tableview.
   * @param event A button click on the add button.
   */
  @FXML
  protected void handleAddButton(ActionEvent event) {
    handleEditButton(event);
  }

  /**
   * Method for handling the editing of a chosen entry in the tableview.
   * @param event A button click on the edit button.
   */
  @FXML
  protected void handleEditButton(ActionEvent event) {
    //Instantiate FXML loader and loads the popup for adding income
    FXMLLoader loader = new FXMLLoader();
    loader.setLocation(getClass().getResource("/view/AddIncome.fxml"));

    Income newIncome = null;
    String dialogTitle = "";
    // Load the FXML file for your dialog box
    Dialog<Income> dialog = new Dialog<>();
    dialog.initModality(Modality.APPLICATION_MODAL);

    try {
      // Set the Dialog's content to the loaded FXML file
      dialog.getDialogPane().setContent(loader.load());
    } catch (IOException e) {
      e.printStackTrace();
    }

    // Get the controller for the loaded FXML file
    AddIncomeController dialogController = loader.getController();

    /**
     * The mode of the dialog. NEW if new contact, EDIT if edit existing contact.
     */
    DialogMode dialogMode;
    //Sets the title of the dialog box
    if (event.getSource().equals(addBtn)) {
      dialogMode = DialogMode.ADD;
      dialogTitle = "Add income";
    }
    else if (event.getSource().equals(editBtn) && incomeTableView.getSelectionModel().getSelectedItem() != null) {
      dialogMode = DialogMode.EDIT;
      dialogTitle = "Edit income";
      //Gets the selected item from the table
      newIncome = incomeTableView.getSelectionModel().getSelectedItem();
      //Binds the selected item to another item which is defined in the ItemController
      dialogController.setIncome(newIncome);
    }
    else {
      return;
    }

    dialog.setTitle(dialogTitle);
    // Show the Dialog and wait for the user to close it
    dialog.showAndWait();
    //Get the newly created income from the dialog pane
    newIncome = dialogController.getNewIncome();

    //Adds the new item to the register
    if (newIncome != null && dialogMode == DialogMode.ADD) {
      incomeRegister.addItem(newIncome);
    }
    //Updates the tableview using the register
    refreshTableView();
  }

  /**
   * Deletes an entry from the tableview, if an entry has been selected. The method brings up a popup window, asking for confirmation for deleting the entry.
   * @param event A button click on the delete button
   */
  @FXML
  public void handleDeleteBtn(ActionEvent event) {
    //Gets the selected item from the tableview
    Income chosenIncome = incomeTableView.getSelectionModel().getSelectedItem();
    //Exits the method if nothing is selected
    if (chosenIncome == null) {
      return;
    }
    //Brings up a confirmation popup
    Optional<ButtonType> isConfirmed = showConfirmationDialog();
    if (isConfirmed.isPresent() && isConfirmed.get() == ButtonType.OK) {
      incomeRegister.removeItem(chosenIncome);
      refreshTableView();
    }
  }

  /**
   * Method for synching the register with the tableview. The observable list to which the tableview is set, is being refilled with all the entries
   * in the register, keeping it updated with new changes.
   */
  protected void refreshTableView() {
    this.income.setAll(incomeRegister.getItems());
    this.sum.setText(String.valueOf(incomeRegister.getTotalSum()));
  }

  /**
   * Returns an optional, which is a popup alert box, asking for confirmation for deleting an entry.
   * @return An alert box, asking for confirmation for deleting the selected entry of the tableview.
   */
  private Optional<ButtonType> showConfirmationDialog() {
    Alert alert = new Alert(AlertType.CONFIRMATION);
    alert.setTitle("Confirm Delete");
    alert.setHeaderText("Delete Confirmation");
    alert.setContentText("Are you sure you would like to delete the selected entry?");

    return alert.showAndWait();
  }

  /**
   * Method that either reads data from a file with which it fills an income register, if older changes exist, or instantiates an income register if the file is empty.
   * @param fileName The name of the file that is being read from.
   * @return An object of type IncomeRegister.
   * @throws IOException If an error occurs while reading from the file.
   */
  public IncomeRegister loadIncomeDataFromFile(String fileName) throws IOException {
    FileHandling fileHandling = new FileHandling();
    //Instantiate new incomeRegister
    if (fileHandling.isEmpty(fileName)) {
      incomeRegister = new IncomeRegister();
    } else { //Load previous income register
      try {
        incomeRegister = fileHandling.readIncomeRegisterFromFile(fileName);
      } catch (IOException e) {
        e.printStackTrace();
      }
    }
    return incomeRegister;
  }

  /**
   * Saves the changes made to the income tableview by writing the information to a file.
   * @param fileName The name of the file that is written to.
   * @throws IOException If an error occurs while writing to the file.
   */
  public void saveDataToFile(String fileName) throws IOException {
    FileHandling fileHandling = new FileHandling();
    fileHandling.writeItemRegisterToFile(incomeRegister, fileName);
  }

  /**
   * Switches scenes from the income scene to another, by loading a new FXML file and setting the scene to this location.
   * The destination depends entirely on which button is pressed.
   * @param event A button click on the buttons on the buttonbar or the next button
   * @throws IOException If an error occurs with loading any of the FXML files.
   */
  @FXML
  public void switchScene(ActionEvent event) throws IOException {
    //Always saving the data when switching scenes
    saveDataToFile("Income");
    FXMLLoader loader = new FXMLLoader();
    if (event.getSource() == expenseBtn) {
      loader.setLocation(SceneController.class.getResource("/view/Expenses.fxml"));
    } else if (event.getSource() == overviewBtn) {
      loader.setLocation(SceneController.class.getResource("/view/Overview.fxml"));
    } else if (event.getSource() == returnBtn) {
      loader.setLocation(SceneController.class.getResource("/view/FirstMenu.fxml"));
    } else if (event.getSource() == budgetBtn) {
      loader.setLocation(SceneController.class.getResource("/view/BudgetNewest.fxml"));
    }
      Parent root = loader.load();
      Stage stage = (Stage) ((Node) event.getSource()).getScene().getWindow();
      Scene scene = new Scene(root);
      stage.setScene(scene);
      stage.show();
    }
  }
